"use strict";
var __defProp = Object.defineProperty;
var __getOwnPropSymbols = Object.getOwnPropertySymbols;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __propIsEnum = Object.prototype.propertyIsEnumerable;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp.call(b, prop))
      __defNormalProp(a, prop, b[prop]);
  if (__getOwnPropSymbols)
    for (var prop of __getOwnPropSymbols(b)) {
      if (__propIsEnum.call(b, prop))
        __defNormalProp(a, prop, b[prop]);
    }
  return a;
};
Object.defineProperties(exports, { __esModule: { value: true }, [Symbol.toStringTag]: { value: "Module" } });
var storePersist = require("./storePersist.js");
var React = require("react");
var fastEquals = require("fast-equals");
require("immer");
function _interopDefaultLegacy(e) {
  return e && typeof e === "object" && "default" in e ? e : { "default": e };
}
var React__default = /* @__PURE__ */ _interopDefaultLegacy(React);
function useEqualityRef(x) {
  const ref = React.useRef(x);
  if (!fastEquals.circularDeepEqual(x, ref.current)) {
    ref.current = x;
  }
  return ref.current;
}
const ignore = () => {
};
function useAction(action, arg, { watchOnly, updateOnMount, clearBeforeUpdate, dormant, throttle } = {}) {
  const [counter, setCounter] = React.useState(0);
  React.useEffect(() => {
    if (updateOnMount && !dormant)
      action.execute(arg, { clearBeforeUpdate }).catch(ignore);
  }, []);
  React.useEffect(() => {
    if (dormant) {
      return;
    }
    if (!watchOnly)
      action.get(arg).catch(ignore);
    return action.subscribe(arg, () => {
      setCounter((c) => c + 1);
      if (!watchOnly)
        action.get(arg).catch(ignore);
    }, { runNow: true, throttle });
  }, [action, useEqualityRef(arg), watchOnly, dormant, throttle]);
  return React.useMemo(() => {
    var _a, _b;
    if (dormant) {
      return [void 0, { error: void 0, isLoading: false }];
    }
    const instance = action.getCache(arg);
    return [
      ((_a = instance == null ? void 0 : instance.current) == null ? void 0 : _a.kind) === "value" ? instance.current.value : void 0,
      { error: ((_b = instance == null ? void 0 : instance.current) == null ? void 0 : _b.kind) === "error" ? instance.current.error : void 0, isLoading: !!(instance == null ? void 0 : instance.inProgress) }
    ];
  }, [action, useEqualityRef(arg), dormant, counter]);
}
class Action extends storePersist.Action {
  useAction(arg, options) {
    return useAction(this, arg, options);
  }
}
function createSelector(stringSelector) {
  const parts = stringSelector.split(".");
  return (state) => parts.reduce((v, p) => v instanceof Object && p in v ? v[p] : void 0, state);
}
function setWithSelector(state, stringSelector, value) {
  const parts = stringSelector.split(".");
  const prefix = parts.slice(0, -1);
  const last = parts[parts.length - 1];
  for (const part of prefix) {
    if (state instanceof Object && part in state) {
      state = state[part];
    } else {
      throw Error(`Could not set ${stringSelector} because path ${stringSelector} doesn't exist`);
    }
  }
  if (state instanceof Object) {
    state[last] = value;
  } else {
    throw Error(`Could not set ${stringSelector} because path ${stringSelector} doesn't exist`);
  }
}
function useStoreState(store, ...args) {
  var _a, _b, _c, _d;
  let selector, deps, options;
  if (args[0] instanceof Function) {
    selector = args[0];
    deps = (_a = args[1]) != null ? _a : [];
    options = (_b = args[2]) != null ? _b : {};
  } else if (typeof args[0] === "string") {
    selector = createSelector(args[0]);
    deps = [args[0]];
    options = (_c = args[1]) != null ? _c : {};
  } else {
    selector = (x) => x;
    deps = [];
    options = (_d = args[0]) != null ? _d : {};
  }
  const [counter, setCounter] = React.useState(0);
  const value = React.useMemo(() => selector(store.getState()), [store, counter, ...deps]);
  React.useEffect(() => store.subscribe(selector, () => setCounter((c) => c + 1), { runNow: true, throttle: options.throttle }), [store, ...deps]);
  return value;
}
function useStoreProp(store, selector) {
  const value = useStoreState(store, selector);
  const update = (value2) => store.update((state) => setWithSelector(state, selector, value2));
  return [value, update];
}
class Store extends storePersist.Store {
  useState(...args) {
    return useStoreState(this, ...args);
  }
  useProp(selector) {
    return useStoreProp(this, selector);
  }
}
class StoreScope {
  constructor(defaultValue) {
    this.defaultValue = defaultValue;
    this.Provider = ({ children, store }) => {
      const _store = React.useMemo(() => store != null ? store : new Store(this.defaultValue), [store]);
      return /* @__PURE__ */ React__default["default"].createElement(this.context.Provider, {
        value: _store
      }, children);
    };
    this.withScope = (Component) => {
      return (args) => /* @__PURE__ */ React__default["default"].createElement(this.Provider, null, /* @__PURE__ */ React__default["default"].createElement(Component, __spreadValues({}, args)));
    };
    this.context = React.createContext(new Store(defaultValue));
  }
  useStore() {
    return React.useContext(this.context);
  }
  useState(...args) {
    const store = this.useStore();
    return useStoreState(store, ...args);
  }
  useProp(selector) {
    const store = this.useStore();
    return useStoreProp(store, selector);
  }
}
exports.StorePersist = storePersist.StorePersist;
exports.Action = Action;
exports.Store = Store;
exports.StoreScope = StoreScope;
exports.useAction = useAction;
exports.useStoreState = useStoreState;
//# sourceMappingURL=react.js.map
