var __defProp = Object.defineProperty;
var __getOwnPropSymbols = Object.getOwnPropertySymbols;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __propIsEnum = Object.prototype.propertyIsEnumerable;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp.call(b, prop))
      __defNormalProp(a, prop, b[prop]);
  if (__getOwnPropSymbols)
    for (var prop of __getOwnPropSymbols(b)) {
      if (__propIsEnum.call(b, prop))
        __defNormalProp(a, prop, b[prop]);
    }
  return a;
};
var __restKey = (key) => typeof key === "symbol" ? key : key + "";
var __objRest = (source, exclude) => {
  var target = {};
  for (var prop in source)
    if (__hasOwnProp.call(source, prop) && exclude.indexOf(prop) < 0)
      target[prop] = source[prop];
  if (source != null && __getOwnPropSymbols)
    for (var prop of __getOwnPropSymbols(source)) {
      if (exclude.indexOf(prop) < 0 && __propIsEnum.call(source, prop))
        target[prop] = source[prop];
    }
  return target;
};
import { freeze, enableMapSet, enablePatches, produce, applyPatches } from "immer";
import { circularDeepEqual } from "fast-equals";
function hash(arg) {
  return JSON.stringify({ arg }, (_key, value) => {
    if (value instanceof Object && !(value instanceof Array)) {
      const obj = {};
      for (const key of Object.keys(value).sort()) {
        obj[key] = value[key];
      }
      return obj;
    }
    return value;
  });
}
const sleep = (ms) => new Promise((r) => setTimeout(r, ms));
async function retry(action, retries, delay = 100) {
  try {
    return await action();
  } catch (e) {
    if (retries <= 0)
      throw e;
    await sleep(delay);
    return retry(action, retries - 1, delay * 10);
  }
}
function throttle(fn, ms) {
  let last = 0;
  let lastArgs;
  let timeout;
  function run() {
    const args = lastArgs;
    last = Date.now();
    lastArgs = void 0;
    timeout = void 0;
    fn(...args);
  }
  return function(...args) {
    const now = Date.now();
    lastArgs = args;
    if (timeout)
      ;
    else if (now < last + ms) {
      timeout = setTimeout(run, last + ms - now);
    } else {
      run();
    }
  };
}
const RESTART_UPDATE = Symbol("RESTART_UPDATE");
class Store {
  constructor(state, options = { log: (...data) => console.error(...data) }) {
    this.state = state;
    this.options = options;
    this.subscriptions = /* @__PURE__ */ new Set();
    this.reactions = /* @__PURE__ */ new Set();
    this.patches = new Array();
    this.inversePatches = new Array();
    this.notifyScheduled = false;
    freeze(state, true);
    enableMapSet();
    enablePatches();
  }
  getState() {
    return this.state;
  }
  update(update) {
    this.checkLock();
    this.state = produce(this.state, (draft) => update(draft, this.state), (patches, inversePatches) => {
      this.patches.push(...patches);
      this.inversePatches.push(...inversePatches);
    });
    this.notify();
  }
  set(state) {
    this.checkLock();
    this.state = produce(this.state, () => state, (patches, inversePatches) => {
      this.patches.push(...patches);
      this.inversePatches.push(...inversePatches);
    });
    this.notify();
  }
  applyPatches(patches) {
    this.checkLock();
    this.state = applyPatches(this.state, patches);
    this.patches.push(...patches);
    this.notify();
  }
  subscribe(selector, listener, { runNow = true, throttle: throttle$1 = 0 } = {}) {
    const throttledListener = throttle$1 ? throttle(listener, throttle$1) : listener;
    let value = selector(this.state);
    const internalListener = (state, _patches, _inversePatches, init) => {
      try {
        const oldValue = value;
        value = selector(state);
        if (!init && circularDeepEqual(value, oldValue))
          return;
        (init ? listener : throttledListener)(value, oldValue, this.state);
      } catch (e) {
        this.options.log("Failed to execute listener:", e);
      }
    };
    this.subscriptions.add(internalListener);
    if (runNow)
      internalListener(this.state, [], [], true);
    return () => {
      this.subscriptions.delete(internalListener);
    };
  }
  addReaction(selector, reaction, { runNow = true } = {}) {
    let value = selector(this.state);
    const internalListener = (init) => {
      let hasChanged = false;
      try {
        this.lock = "reaction";
        const oldValue = value;
        value = selector(this.state);
        if (!init && circularDeepEqual(value, oldValue))
          return;
        this.state = produce(this.state, (draft) => reaction(value, draft, this.state, oldValue), (patches, inversePatches) => {
          hasChanged = true;
          this.patches.push(...patches);
          this.inversePatches.push(...inversePatches);
        });
      } catch (e) {
        this.options.log("Failed to execute reaction:", e);
      } finally {
        delete this.lock;
      }
      if (hasChanged && !init)
        throw RESTART_UPDATE;
      if (hasChanged && init)
        this.notify();
    };
    this.reactions.add(internalListener);
    if (runNow)
      internalListener(true);
    return () => {
      this.reactions.delete(internalListener);
    };
  }
  subscribePatches(listener) {
    const internalListener = (_state, patches, inversePatches) => {
      try {
        listener(patches, inversePatches);
      } catch (e) {
        this.options.log("Failed to execute patch listener:", e);
      }
    };
    this.subscriptions.add(internalListener);
    return () => {
      this.subscriptions.delete(internalListener);
    };
  }
  checkLock() {
    if (this.lock === "reaction") {
      throw Error("You cannot call update from within a reaction. Use the passed draft instead.");
    }
  }
  notify() {
    try {
      for (const reaction of this.reactions) {
        reaction();
      }
    } catch (e) {
      return this.notify();
    }
    if (this.notifyScheduled)
      return;
    this.notifyScheduled = true;
    (async () => {
      await Promise.resolve();
      this.notifyScheduled = false;
      const patches = this.patches;
      const inversePatches = this.inversePatches;
      this.patches = [];
      this.inversePatches = [];
      for (const subscription of this.subscriptions) {
        subscription(this.state, patches, inversePatches);
      }
    })();
  }
}
const _Action = class {
  constructor(action, options = {}) {
    this.action = action;
    this.options = options;
    this.cache = new Store(/* @__PURE__ */ new Map());
    _Action.allActions.push(this);
    enableMapSet();
  }
  static clearCacheAll() {
    for (const action of _Action.allActions) {
      action.clearCacheAll();
    }
  }
  getCache(arg) {
    const key = hash(arg);
    return this.cache.getState().get(key);
  }
  getCacheValue(arg) {
    var _a;
    const instance = this.getCache(arg);
    return ((_a = instance == null ? void 0 : instance.current) == null ? void 0 : _a.kind) === "value" ? instance.current.value : void 0;
  }
  getCacheError(arg) {
    var _a;
    const instance = this.getCache(arg);
    return ((_a = instance == null ? void 0 : instance.current) == null ? void 0 : _a.kind) === "error" ? instance.current.error : void 0;
  }
  setCache(arg, value) {
    this.updateInstance(arg, (instance) => {
      instance.current = { kind: "value", value, t: new Date() };
      delete instance.inProgress;
      instance.invalid = false;
    });
    this.setTimers(arg);
  }
  updateCache(arg, update) {
    let found = false;
    this.updateInstance(arg, (instance) => {
      var _a;
      if (((_a = instance.current) == null ? void 0 : _a.kind) === "value") {
        update(instance.current.value);
        instance.invalid = false;
        found = true;
      }
    });
    if (found) {
      this.setTimers(arg);
    }
    return found;
  }
  updateCacheAll(update) {
    var _a;
    this.cache.update((state) => {
      var _a2;
      for (const instance of state.values()) {
        if (((_a2 = instance.current) == null ? void 0 : _a2.kind) === "value") {
          update(instance.current.value, instance.arg);
          instance.invalid = false;
        }
      }
    });
    for (const instance of this.cache.getState().values()) {
      if (((_a = instance.current) == null ? void 0 : _a.kind) === "value") {
        this.setTimers(instance.arg);
      }
    }
  }
  clearCache(arg) {
    const key = hash(arg);
    this.cache.update((state) => {
      const instance = state.get(key);
      if (instance == null ? void 0 : instance.invalidateTimer) {
        clearTimeout(instance.invalidateTimer);
      }
      if (instance == null ? void 0 : instance.clearTimer) {
        clearTimeout(instance.clearTimer);
      }
      state.delete(key);
    });
  }
  clearCacheAll() {
    for (const { arg } of this.cache.getState().values()) {
      this.clearCache(arg);
    }
  }
  invalidateCache(arg) {
    const key = hash(arg);
    this.cache.update((state) => {
      const instance = state.get(key);
      if (instance) {
        delete instance.inProgress;
        instance.invalid = true;
      }
      if (instance == null ? void 0 : instance.invalidateTimer) {
        clearTimeout(instance.invalidateTimer);
        delete instance.invalidateTimer;
      }
    });
  }
  invalidateCacheAll() {
    for (const { arg } of this.cache.getState().values()) {
      this.invalidateCache(arg);
    }
  }
  async get(arg, { clearBeforeUpdate = false, retries = 0 } = {}) {
    const key = hash(arg);
    const fromCache = this.cache.getState().get(key);
    if ((fromCache == null ? void 0 : fromCache.current) && !fromCache.invalid) {
      if (fromCache.current.kind === "value")
        return fromCache.current.value;
      else
        throw fromCache.current.error;
    }
    if (fromCache == null ? void 0 : fromCache.inProgress) {
      return fromCache.inProgress;
    }
    return this.execute(arg, { clearBeforeUpdate, retries });
  }
  async execute(arg, { clearBeforeUpdate = false, retries = 0 } = {}) {
    var _a, _b;
    const key = hash(arg);
    const task = retry(() => this.action(arg), retries);
    this.updateInstance(arg, (instance) => {
      if (clearBeforeUpdate)
        delete instance.current;
      instance.inProgress = task;
    });
    try {
      const value = await task;
      if (task === ((_a = this.cache.getState().get(key)) == null ? void 0 : _a.inProgress)) {
        this.setCache(arg, value);
      }
      return value;
    } catch (e) {
      if (task === ((_b = this.cache.getState().get(key)) == null ? void 0 : _b.inProgress)) {
        this.updateInstance(arg, (instance) => {
          instance.current = { kind: "error", error: e, t: new Date() };
          delete instance.inProgress;
          instance.invalid = false;
        });
        this.setTimers(arg);
      }
      throw e;
    }
  }
  setTimers(arg) {
    var _a, _b, _c, _d;
    const instance = this.updateInstance(arg, (instance2) => {
      if (instance2.invalidateTimer) {
        clearTimeout(instance2.invalidateTimer);
        delete instance2.invalidateTimer;
      }
      if (instance2.clearTimer) {
        clearTimeout(instance2.clearTimer);
        delete instance2.clearTimer;
      }
    });
    let { invalidateAfter = _Action.options.invalidateAfter, clearAfter = _Action.options.clearAfter } = this.options;
    if (invalidateAfter instanceof Function) {
      invalidateAfter = invalidateAfter(((_a = instance.current) == null ? void 0 : _a.kind) === "value" ? instance.current.value : void 0, ((_b = instance.current) == null ? void 0 : _b.kind) === "error" ? instance.current.error : void 0);
    }
    if (clearAfter instanceof Function) {
      clearAfter = clearAfter(((_c = instance.current) == null ? void 0 : _c.kind) === "value" ? instance.current.value : void 0, ((_d = instance.current) == null ? void 0 : _d.kind) === "error" ? instance.current.error : void 0);
    }
    let invalidateTimer, clearTimer;
    if (invalidateAfter !== void 0 && invalidateAfter !== Infinity) {
      invalidateTimer = setTimeout(() => {
        this.invalidateCache(arg);
      }, invalidateAfter);
    }
    if (clearAfter !== void 0 && clearAfter !== Infinity) {
      clearTimer = setTimeout(() => {
        this.clearCache(arg);
      }, clearAfter);
    }
    this.updateInstance(arg, (instance2) => {
      instance2.invalidateTimer = invalidateTimer;
      instance2.clearTimer = clearTimer;
    });
  }
  subscribe(arg, listener, { runNow, throttle: throttle2 } = {}) {
    const key = hash(arg);
    return this.cache.subscribe((state) => {
      var _a;
      return (_a = state.get(key)) != null ? _a : { arg };
    }, (instance) => {
      var _a, _b;
      return listener(((_a = instance.current) == null ? void 0 : _a.kind) === "value" ? instance.current.value : void 0, {
        error: ((_b = instance.current) == null ? void 0 : _b.kind) === "error" ? instance.current.error : void 0,
        isLoading: !!instance.inProgress
      }, instance);
    }, { runNow, throttle: throttle2 });
  }
  updateInstance(arg, update) {
    const key = hash(arg);
    this.cache.update((state) => {
      let instance = state.get(key);
      if (!instance) {
        instance = { arg };
        state.set(key, instance);
      }
      update(instance);
    });
    return this.cache.getState().get(key);
  }
};
let Action = _Action;
Action.allActions = new Array();
Action.options = {};
const isAncestor = (ancestor, path) => {
  return ancestor.length <= path.length && ancestor.every((v, i) => v === "*" || path[i] === "*" || String(v) === String(path[i]));
};
const get = (value, path) => {
  for (const part of path) {
    if (value instanceof Object && part in value)
      value = value[part];
    else {
      value = void 0;
      break;
    }
  }
  return value;
};
const split = (value, path) => {
  const [first, ...rest] = path;
  if (first === void 0)
    return [value, []];
  if (rest.length === 0) {
    if (first === "*")
      return [{}, Object.entries(value).map(([k, v]) => ({ path: [k], value: v }))];
    if (!(first in value))
      return [value, []];
    const _a = value, { [first]: subValue } = _a, newValue2 = __objRest(_a, [__restKey(first)]);
    return [newValue2, [{ path: [first], value: subValue }]];
  }
  const newValue = __spreadValues({}, value);
  const subValues = new Array();
  for (const key of first === "*" ? Object.keys(value) : [first]) {
    if (!(newValue[key] instanceof Object))
      return [value, []];
    const result = split(newValue[key], rest);
    newValue[key] = result[0];
    subValues.push(...result[1].map((s) => ({ path: [key, ...s.path], value: s.value })));
  }
  return [newValue, subValues];
};
class StorePersist {
  constructor(store, storage, options = {}) {
    this.store = store;
    this.storage = storage;
    this.options = options;
    this.initialization = this.load();
    this.isStopped = false;
    this.saveQueue = new Array();
    this.isSaving = false;
    if (!options.paths) {
      this.paths = [{ path: [], throttleMs: options.throttleMs }];
    } else {
      this.paths = options.paths.map((p) => {
        var _a;
        return {
          path: (typeof p === "object" ? p.path : p).split("."),
          throttleMs: (_a = typeof p === "object" ? p.throttleMs : void 0) != null ? _a : options.throttleMs
        };
      });
    }
    this.paths.sort((a, b) => b.path.length - a.path.length);
  }
  async load() {
    const storage = this.storage;
    let keys;
    if ("keys" in storage) {
      keys = await storage.keys();
    } else {
      const length = storage.length instanceof Function ? await storage.length() : storage.length;
      keys = new Array();
      for (let i = 0; i < length; i++) {
        const key = await storage.key(i);
        if (key !== null)
          keys.push(key);
      }
    }
    keys.sort((a, b) => a.length - b.length);
    const patches = new Array();
    for (const key of keys) {
      const value = await storage.getItem(key);
      if (value) {
        patches.push({
          path: key === "" ? [] : key.split("."),
          op: "replace",
          value: value === "undefined" ? void 0 : parse(value),
          persist: this
        });
      }
    }
    if (this.isStopped)
      return;
    if (patches.length > 0)
      this.store.applyPatches(patches);
    this.watch(patches.length === 0);
  }
  watch(isInitialized) {
    this.sub = this.store.subscribePatches((patches) => {
      for (const patch of patches) {
        if (patch.persist === this) {
          isInitialized = true;
        } else if (isInitialized) {
          this.addToSaveQueue(patch.path);
        }
      }
    });
  }
  addToSaveQueue(path) {
    var _a;
    const match = this.paths.find((p) => isAncestor(p.path, path));
    if (!match)
      return;
    path = path.slice(0, match.path.length);
    const t = Date.now() + ((_a = match.throttleMs) != null ? _a : 0);
    if (this.saveQueue.some((i) => i.t <= t && isAncestor(i.path, path)))
      return;
    this.saveQueue.push({ path, t });
    this.saveQueue.sort((a, b) => a.t - b.t);
    this.run();
  }
  async run() {
    if (this.saveTimeout)
      clearTimeout(this.saveTimeout);
    const next = this.saveQueue[0];
    if (!next)
      return;
    if (next.t > Date.now()) {
      this.saveTimeout = setTimeout(() => this.run(), next.t - Date.now());
      return;
    }
    try {
      this.isSaving = true;
      this.saveQueue.shift();
      await this.save(next.path);
    } finally {
      this.isSaving = false;
      this.run();
    }
  }
  async save(path) {
    var _a, _b;
    const { store, storage } = this;
    let value = get(store.getState(), path);
    const subPaths = this.paths.filter((p) => isAncestor(path, p.path) && p.path.length > path.length).map((p) => p.path.slice(path.length));
    const subValues = new Array();
    for (const subPath of subPaths) {
      const result2 = split(value, subPath);
      value = result2[0];
      subValues.push(...result2[1].map((s) => ({ path: [...path, ...s.path], value: s.value })));
    }
    const result = storage.setItem(path.join("."), (_a = stringify(value)) != null ? _a : "undefined");
    if (result instanceof Promise)
      await result;
    for (const { path: path2, value: value2 } of subValues) {
      const result2 = storage.setItem(path2.join("."), (_b = stringify(value2)) != null ? _b : "undefined");
      if (result2 instanceof Promise)
        await result2;
    }
  }
  stop() {
    var _a;
    this.isStopped = true;
    (_a = this.sub) == null ? void 0 : _a.call(this);
  }
}
function stringify(value) {
  function prepare(value2) {
    if (value2 instanceof Date) {
      return { __date: value2.toJSON() };
    }
    if (value2 instanceof Set) {
      return { __set: Array.from(value2).map(prepare) };
    }
    if (value2 instanceof Map) {
      return { __map: Array.from(value2.entries()).map((entry) => entry.map(prepare)) };
    }
    if (value2 instanceof Array) {
      return value2.map(prepare);
    }
    if (value2 instanceof Object) {
      return Object.fromEntries(Object.entries(value2).map(([k, v]) => [k, prepare(v)]));
    }
    return value2;
  }
  return JSON.stringify(prepare(value));
}
function parse(value) {
  return JSON.parse(value, (_key, value2) => {
    if (value2 instanceof Object && "__date" in value2) {
      return new Date(value2.__date);
    }
    if (value2 instanceof Object && "__set" in value2) {
      return new Set(value2.__set);
    }
    if (value2 instanceof Object && "__map" in value2) {
      return new Map(value2.__map);
    }
    return value2;
  });
}
export { Action as A, StorePersist as S, Store as a };
//# sourceMappingURL=storePersist.js.map
